
from .consts import *
from .cltrack import ClTrack
from .dltrack import DlTrack
from .track import Track
import Live
import math
from .scene import Scene
from .clip import Clip
from .mute import Mute
from .solo import Solo
from collections import defaultdict
from .state import *


class Actions:

    def __init__(self, parent, tracks, song, state):
        self.tracks = tracks
        self.state = state
        self.song = song
        self.__parent = parent
        self.timerCounter = 0
        self.quantize_timer = Live.Base.Timer(callback=self.on_quantize_changed_timer_callback, interval=1, repeat=False)
        self.record_timer = Live.Base.Timer(callback=self.record_timer_callback, interval=1, repeat=False)
        self.current_time_change_timer = Live.Base.Timer(callback=self.current_time_change_timer_callback, interval=1, repeat=False)
        self.should_start_tracks = False
        self.playing_clips = []
        self.queued_scene = -1
        self.scenes = []
        self.clips = []
        self.mutes = []
        self.solo = Solo(self, self.song, self.tracks)
        self.song.add_scenes_listener(self.on_scene_change)
        self.lastQuantization = -1
        self.ignore_release_command = False
        self.press_state = -1

    ##### HELPER FUNCTIONS #####

    @staticmethod
    def get_looper_type(type):
        types = {
            0: DlTrack,
            1: ClTrack,
            2: Track
        }
        return types.get(type)

    def update_tracks(self, tracks):
        # self.__parent.send_message("updating tracks")
        self.tracks = tracks

    def call_method_on_tracks(self,vendor, track_num, track_type, method_name, *args):
        tracks = self.tracks.get(str(track_num))
        self.send_message("call method on track num: " + str(track_num) + " is tracks none? " + str(tracks is None))

        if tracks is not None:
            for track in tracks:
                if self.press_state == PressState.PRESS and (track.lastState == RECORDING_STATE or track.lastState == OVERDUB_STATE):
                    self.__parent.send_message("ignoring next release")
                    self.ignore_release_command = True

                self.__parent.send_message("attempting to call method " + method_name + " on track name: " + track.track.name + " track num: " + str(track.trackNum) + " button:" + str(track.button_num))

                #checks if track is correct type (CL or DL) and if it's linked to this pedal instance or is requesting to be linked
                if isinstance(track, self.get_looper_type(track_type)) and (track.button_num != -1 or (method_name == 'set_group_master' or method_name == 'link_button' or method_name == 'remove_clip_slot' or vendor != DATALOOPER_VENDOR)):
                    self.__parent.send_message("calling method " + method_name + " on track name: " + track.track.name + " track num: " + str(track.trackNum))
                    obj = getattr(track, method_name,"No method found")
                    self.__parent.send_message("getattr returned:" + str(obj))
                    obj(*args)

    def call_method_on_all_tracks(self, track_type, method_name, *args):
        self.send_message("call method on all tracks")

        for tracks in self.tracks.values():
            for track in tracks:
                if isinstance(track, self.get_looper_type(track_type)):
                    # ensures that all clips get cleared if multiple CL#(same number) tracks.
                    if method_name == 'clear' :
                        track.should_clear_all = True
                    self.send_message("calling method" + method_name + " on track: " + track.track.name)
                    getattr(track, method_name)(*args)

    def check_uniform_state(self,track_type, state):
        for tracks in self.tracks.values():
            for track in tracks:
                # self.send_message("track " + str(track.track.name) + " State:" + str(track.lastState))
                if isinstance(track, self.get_looper_type(track_type)) and track.lastState not in state:
                    return False
        return True

    def check_track_clear(self, tracks):
        if tracks is None:
            self.__parent.send_message("no tracks present")
            return True
        for track in tracks:
            if track.lastState != CLEAR_STATE:
                return False
        return True

    ##### TRACK ACTIONS #####

    def new_clip(self, data):
        self.call_method_on_tracks(data.vendor,data.data[0]-1, CLIP_TRACK, "new_clip")

    def looper_control(self, data):
        self.send_message("looper control: " + str(data.data[1]) + "looper type: " + str(data.data[2]) + " data4: " + str(data.data[3]) + " press state: " + str(data.data[4]))
        #on_all indica tes if action should be on tracks that are not record armed.
        #if calling clear on all cl tracks, clip should clear even if not record armed, if calling clear on a single cl track, it should not execute, hence this param.

        looper_num = data.data[0] - 1
        looper_action = data.data[1]
        looper_type = data.data[2]
        self.press_state = data.data[4]
        action = LOOPER_ACTIONS.get(data.data[1])

        #if it's an exclusive action, it needs to be triggered on all tracks, to ensure the other tracks stop
        if looper_action == LooperActions.EXCLUSIVE_RECORD_IMMEDIATELY or looper_action == LooperActions.EXCLUSIVE_RECORD_QUANTIZED:
            self.send_message("Exclusive rec: looper number: " + str(looper_num))
            self.call_method_on_all_tracks(looper_type, action, looper_num)
        elif looper_action in LooperActions.FADEABLE_ACTIONS or looper_action in LooperActions.RECORD_ACTIONS or looper_action == LooperActions.NONE and not (self.press_state == PressState.RELEASE and self.ignore_release_command):
            self.call_method_on_tracks(data.vendor,looper_num, looper_type, action, data.data[3]) if looper_num != ALL_LOOPERS else self.call_method_on_all_tracks(looper_type, action, data.data[3])
        else:
            self.call_method_on_tracks(data.vendor,looper_num, looper_type, action) if looper_num != ALL_LOOPERS else self.call_method_on_all_tracks(looper_type, action)

        if self.press_state == PressState.RELEASE:
            self.__parent.send_message("resetting release ignore")
            self.ignore_release_command = False

    def clip_control(self, data):
        # 0, 1, 2 || 4, 5, 6 || 8, 9, 10
        trackNum = data.data[0] - 1
        sceneNum = data.data[2] - 1
        self.__parent.send_message("clip control called")
        self.__parent.send_message("trackNum: " + str(trackNum) + " sceneNum: " + str(sceneNum))
        for clip in self.clips:
            if clip.trackNum == trackNum and clip.sceneNum == sceneNum:
                action = CLIP_ACTIONS.get(data.data[1])
                self.__parent.send_message("calling clip action " + str(action))
                getattr(clip, action)(data)

    def scene_control(self, data):
        for scene_control in self.scenes:
            if scene_control.sceneNum == data.data[0]-1:
                if scene_control.sceneNum >= 0 and scene_control.sceneNum < 125:
                    scene_control.fire_scene(data.data[1])
                elif scene_control.sceneNum == 125:
                    scene_control.next_scene()
                elif scene_control.sceneNum == 126:
                    scene_control.prev_scene()

    def change_instance(self, data):
        self.send_message("changing instance to dl preset #" + str(data.data[0]))
        self.send_sysex(DLCommands.KEEP_ALIVE, DLActions.CHANGE_INSTANCE, data.data[0])
        self.send_message(str(data))
        for scene in self.scenes:
            scene.remove()
            scene.button_num = -1
        self.scenes = []

        for clip in self.clips:
            clip.remove()
            clip.button_num = -1
        self.clips = []

        for tracks in self.tracks.values():
            for track in tracks:
                track.button_num = -1

        for mute in self.mutes:
            mute.remove()
            mute.button_num = -1
        self.mutes = []

    def request_midi_map_rebuild(self, data):
        self.send_message("midi map rebuild requested")
        self.__parent.request_rebuild_midi_map()

    def request_state(self, data):
        # data1 = buttonNum
        # data2 = command
        # data3 = data1
        # data4 = data2, etc...

        method = self.__parent.get_method(data.data[1])
        if method == "scene_control":
            linkedScene = self.is_scene_linked(data.data[2]-1, data.data[0])
            if not linkedScene:
                linkedScene = Scene(data.data[2]-1, data.data[0], self.song, self.state, self)
                self.scenes.append(linkedScene)
            linkedScene.request_state()
        elif method == "clip_control":
            # self.send_message("requesting state on button number: " + str(data.data[0]) + " track #" + str(data.data[2]-1) + " clip #" + str(data.data[4]-1))
            linkedClip = self.is_clip_linked(data.data[2]-1, data.data[4]-1)
            if not linkedClip:
                linkedClip = Clip(data.data[2]-1, data.data[4]-1, data.data[0], self.song, self.state, self)
                self.clips.append(linkedClip)
            linkedClip.request_state()
        elif method == "looper_control":
            self.call_method_on_tracks(data.vendor,data.data[2]-1, data.data[4], "link_button", data.data[0], LOOPER_ACTIONS.get(data.data[3]))
            self.send_message("looper control link button:" + str(data.data[0]))
        elif method == "mute_control":
            # self.send_message("found mute control mute type:" + str(data.data[2]))
            mute = self.is_mute_linked(data)
            if not mute:
                mute = Mute(self, data.data[0], data.data[2], data.data[3], self.song, self.tracks, data.data[4]-1)
                self.mutes.append(mute)
            mute.request_state()
        elif method == "solo_control" :
            self.solo.link_button(data.data[0], data.data[2]-1)

    def is_mute_linked(self, data):
        for mute in self.mutes:
            if mute.button_num == data.data[0] and data.data[1] == mute.mute_type and data.data[2] == mute.mute_what:
                return mute
        return False

    def is_clip_linked(self, trackNum, sceneNum):
        for clip in self.clips:
            if clip.trackNum == trackNum and clip.sceneNum == sceneNum:
                return clip

        return False

    def is_scene_linked(self, sceneNum, buttonNum):
        for scene in self.scenes:
            if scene.sceneNum == sceneNum and scene.button_num == buttonNum:
                return scene

        return False

    def get_color_array(self, color):
        if color is not None:
            blue = color & 255
            green = (color >> 8) & 255
            red = (color >> 16) & 255
            return red >> 1, 1 & red, green >> 1, 1 & green, blue >> 1, 1 & blue
        else:
            return None

    def smart_clear(self,data):
        for tracks in self.tracks.values():
            for track in tracks:
                if track.group_number != -1 and track.track.is_grouped and track.track.group_track.mute and track.lastState == PLAYING_STATE and not track.track.mute:
                    track.clear_immediately()

    ##### EFFECT ALL DATALOOPER TRACKS ACTIONS #####

    def disable_quantization(self):
        if not self.song.clip_trigger_quantization_has_listener(self.on_quantize_changed):
            self.song.add_clip_trigger_quantization_listener(self.on_quantize_changed)
            self.lastQuantization = self.song.clip_trigger_quantization
            self.song.clip_trigger_quantization = Live.Song.Quantization.q_no_q

    def on_quantize_changed(self):
        self.send_message("quantization disabled")
        self.quantize_timer.start()

    def on_quantize_changed_timer_callback(self):
        self.song.remove_clip_trigger_quantization_listener(self.on_quantize_changed)
        self.call_method_on_all_tracks(CLIP_TRACK, "on_quantize_disabled")
        self.song.clip_trigger_quantization = self.lastQuantization

    def toggle_stop_start(self, data):
        quantizedStop = data.data[2] if data.data[2] else data.data[1]

        if data.data[0] == 4: #all playing clips and dl loopers
            track_type = BOTH_TRACK_TYPES
        elif data.data[0] == 3: #all playing clips
            track_type = 1
        else:
            track_type = data.data[0]

        self.state.should_record = self.song.record_mode
        self.__parent.send_message("toggling stop/start")
        self.__parent.send_message("uniform clear?: " + str(self.check_uniform_state(track_type, [CLEAR_STATE])) + " uniform stop or clear state: " + str(self.check_uniform_state(track_type,[STOP_STATE, CLEAR_STATE])))
        if not self.check_uniform_state(track_type, [CLEAR_STATE]) and self.check_uniform_state(track_type, [STOP_STATE, CLEAR_STATE]):
            self.__parent.send_message("toggling start, quantized? " + str(data.data[1]))
            if data.data[1] == 0:
                self.state.bpm = self.song.tempo
                self.jump_to_next_bar()
                self.should_start_tracks = True
            else:
                self.call_method_on_all_tracks(track_type, "start", data.data[1])
            if self.playing_clips:
                for clip in self.playing_clips:
                    clip.fire()
                self.playing_clips = []

        else:
            self.__parent.send_message("toggling stop")
            self.call_method_on_all_tracks(track_type, "stop", quantizedStop)
            if data.data[0] == 4 or data.data[0] == 3:
                for track in self.song.tracks:
                    if track.playing_slot_index > -1 and track.clip_slots[track.playing_slot_index].has_clip and not track.clip_slots[track.playing_slot_index].is_triggered  :
                        track.clip_slots[track.playing_slot_index].clip.stop()
                        self.playing_clips.append(track.clip_slots[track.playing_slot_index])
                # self.song.stop_all_clips()

    def new_clips_on_all(self, data):
        self.call_method_on_all_tracks(data, CLIP_TRACK, "new_clip")

    def mute_control(self, data):
        for mute in self.mutes:
            if mute.mute_type is data.data[0] and mute.mute_what is data.data[1] and mute.trackNum is data.data[2]-1:
                mute.execute_mute()

    ##### EFFECT ENTIRE SESSION #####

    def on_scene_change(self):
        self.__parent.send_message("scene changed")
        for x, scene in enumerate(self.scenes):
            scene.init_listener()

    def move_session_highlight(self, data):
        self.__parent.send_message("moving session highlight:" + str(data.data[0]))
        if data.data[0] == 0 and self.state.sceneOffset > 0:
            self.state.sceneOffset -= 1
        elif data.data[0] == 1:
            self.state.sceneOffset += 1
        elif data.data[0] == 2 and self.state.trackOffset > 0:
            self.state.trackOffset -= 1
        elif data.data[0] == 3:
            self.state.trackOffset += 1
        for clip in self.clips:
            clip.remove()
            clip.init_listener()
            clip.update_color()
        self.__parent._set_session_highlight(self.state.trackOffset, self.state.sceneOffset, 3, 3, False)

    def stop_all_playing_clips(self, data):
        self.song.stop_all_clips()
        self.stop_all(data)

    def create_scene(self, data):
        self.song.create_scene(-1)

    def metronome_control(self, data):
        if data.data[0] is 0:
            self.song.metronome = 0
        elif data.data[0] is 1:
            self.song.metronome = 1
        elif data.data[0] is 2:
            self.song.metronome = not self.state.metro
            self.state.metro = not self.state.metro

    def transport_control(self, data):
        if data.data[0] is 0:
            self.song.stop_playing()
        elif data.data[0] is 1:
            self.song.start_playing()
        elif data.data[0] is 2:
            if self.song.is_playing:
                self.song.stop_playing()
            else:
                self.song.start_playing()

    def tap_tempo(self, data):
        self.song.tap_tempo()
        if data.data[0]:
            self.send_message(str(self.state.tap_tempo_counter))
            if self.state.tap_tempo_counter is 0:
                self.state.metro = self.song.metronome
                self.state.ignoreMetroCallback = True
                self.song.metronome = 0
            if self.state.tap_tempo_counter >= data.data[1]:
                self.state.restore_metro()
            elif self.state.metro is not False:
                self.state.tap_tempo_counter += 1

    def jump_to_next_bar(self):
        self.send_message("jumping to next bar")
        self.state.was_recording = self.song.record_mode

        if self.song.record_mode:
            if not self.song.record_mode_has_listener(self.on_record_mode_changed):
                self.song.add_record_mode_listener(self.on_record_mode_changed)
            self.song.record_mode = 0
        elif self.state.should_record:
            self.record_timer_callback()
        else:
            self.set_tempo_jump_time()

    def song_timer_callback(self):
        if self.song.current_song_time_has_listener(self.song_timer_callback):
            self.song.remove_current_song_time_listener(self.song_timer_callback)
        self.send_message("song timer listener callback")
        self.current_time_change_timer.start()

    def current_time_change_timer_callback(self):
        self.song.record_mode = True

    def on_record_mode_changed(self):
        self.song.remove_record_mode_listener(self.on_record_mode_changed)
        self.record_timer.start()

    def record_timer_callback(self):
        if not self.song.current_song_time_has_listener(self.song_timer_callback):
            self.song.add_current_song_time_listener(self.song_timer_callback)
        self.set_tempo_jump_time()

    def set_tempo_jump_time(self):
        self.song.tempo = self.state.bpm
        time = int(self.song.current_song_time) + (self.song.signature_denominator - (
                int(self.song.current_song_time) % self.song.signature_denominator))
        self.send_message("song time: " + str(time))

        if self.state.queued is not False:
            self.state.queued.request_control(MASTER_CONTROL)
            self.state.queued = False
            # if self.state.mode == NEW_SESSION_MODE:
            self.change_mode()
            self.__parent.send_sysex(DLCommands.CHANGE_MODE, 0)
        if self.should_start_tracks:
            self.call_method_on_all_tracks(BOTH_TRACK_TYPES, "start", False)
            self.should_start_tracks = False
        self.song.current_song_time = time - (self.state.bpm / 60000 * 20)

    def change_mode(self, data=False):
        if not data:
            mode = LOOPER_MODE
        else:
            mode = data.data[0]
        self.state.change_mode(self.__parent, mode)

        self.__parent.send_message("mode change to: " + str(mode))

        if mode == NEW_SESSION_MODE:
            if self.song.record_mode:
                self.state.should_record = self.song.record_mode
                self.song.record_mode = False
            self.call_method_on_all_tracks(BOTH_TRACK_TYPES,"change_mode")
        elif mode == LOOPER_MODE:
            self.call_method_on_all_tracks(BOTH_TRACK_TYPES,"change_mode")

        self.state.ignore_tempo_control = False

    def send_sysex(self, *data):
        self.__parent.send_sysex(*data)

    def solo_control(self, data):
        self.send_message("executing solo control")
        self.solo.execute_solo(data.data[0] - 1, data.data[1], data.data[2])

    # def execute_tempo_change(self):
    # # kills timer after 50ms just in case it wants to run forever for some reason
    #     self.timerCounter += 1
    #     if self.song.tempo != self.state.bpm:
    #         self.song.tempo = self.state.bpm
    #         self.tempo_change_timer.stop()
    #     elif self.timerCounter > 50:
    #         self.tempo_change_timer.stop()

    def send_message(self, m):
        self.__parent.send_message(m)

    def start_recording(self, data):
        self.send_message("start recording sysex received")
        self.song.record_mode = True

    def set_record_delay(self, data):
        self.send_message("record delay set to " + str(data.data[2]) + " should persist = " + str(data.data[1]) + " enabled = " + str(data.data[0]))
        RecBeatDelay.enabled = data.data[0]
        RecBeatDelay.should_persist = data.data[1]
        RecBeatDelay.num_beats_delayed = data.data[2]

    def set_record_length(self, data):
        self.send_message("record length set to " + str(data.data[2]) + " should persist = " + str(data.data[1]) + " enabled = " + str(data.data[0]))
        FixedLengthLoop.enabled = data.data[0]
        FixedLengthLoop.should_persist = data.data[1]
        FixedLengthLoop.num_beats_to_loop = data.data[2]

    def launch_song_part(self, data):
        part = data.data[0]
        self.send_message("launching song part from DL Script: " + str(part))
        if LiveOSCBridge.script is not None:
            LiveOSCBridge.script.launch_song_part(data.data[0], data.data[1]-1)
        else:
            self.send_message("liveOSC script none")

    def load_song(self, data):
        pass

    def toggle_groups(self,groups,shouldMute):
        for tracks in self.tracks.values():
            for track in tracks:
                if track.group_number is not -1:
                    self.send_message("toggling group #" + str(track.group_number) + " checking groups " + str(groups))
                    if track.group_number in groups and track.track.is_grouped:
                        track.track.group_track.mute = shouldMute

    def set_group_masters(self,data):
        track_num = data.data[0] - 1
        masters = []
        self.send_message("in actions, attempting setting of group masters on track number " + str(track_num))
        for i in range(4,len(data.asBytes)):
            if data.asBytes[i] > 0 and data.asBytes[i] < 99:
                masters.append(data.asBytes[i])
        self.call_method_on_tracks(data.vendor,track_num, LOOPER_TRACK, 'set_group_master', masters)
