from .cltrack import ClTrack
import Live
from .consts import *

### Controls ###
# 1. Record : Will record a new clip
# 2. Stop : Will stop the recorded clip
# 3. Undo : Since undo is not possible, this will function as a play button
# 4. Clear : Clear will clear the current clip
#################
class ClAudioTrack(ClTrack):
    def __init__(self, parent, track, trackNum, song, state, action_handler):
        super(ClAudioTrack, self).__init__(parent, track, trackNum, song, state, action_handler)
        self.__parent = parent
        self.overdub_track = False
        self.ovr_timer = Live.Base.Timer(callback=self.copy_clip, interval=1, repeat=False)
        self.clr_timer = Live.Base.Timer(callback=self.clear_clip, interval=1, repeat=False)


    ### SCENARIOS ###
    # 1. Clip is empty, recording starts
    # 2. Clip is already in memory and recording, meaning clip will play
    # 3. Clip is playing, new clip found and recording starts
    #################
    def record(self, quantized):
        self.__parent.send_message("record pressed")

        if self.clipSlot == -1:
            self.trigger_record(quantized)            
        elif not self.clipSlot.is_playing or self.clipSlot.is_recording:
            self.fire_clip(quantized)


    def trigger_record(self, quantized = True):
        self.get_new_clip_slot(False)
        self.fire_clip(quantized)

    def undo(self):
        self.get_new_clip_slot(False)

    def link_overdub(self, overdub_track):
        self.overdub_track = overdub_track

    def on_slot_fired(self):
        super(ClAudioTrack, self).on_slot_fired()
        self.send_message("claudiotrack on slot fired")
        if self.track.playing_slot_index >= 0 and self.clipSlot != -1 and self.clipSlot.has_clip and not self.clipSlot.clip.is_recording and self.overdub_track != False:
            if self.clipSlot.clip.is_playing and not self.clipStopping:
                self.send_message("claudiotrack slot fired")
                self.send_message("overdub track: " + str(self.overdub_track))

                self.send_message("starting overdub timer")
                self.ovr_timer.start()
        if self.clipSlot != -1 and self.clipSlot.has_clip and not self.clipSlot.clip.is_recording and self.overdub_track != False and not self.clipSlot.clip.is_playing:
            self.send_message("attempting clear clip")
            self.send_message("overdub track: " + str(self.overdub_track))
            self.clr_timer.start()


    def clear_immediately(self):
        super(ClAudioTrack,self).clear_immediately()
        if self.overdub_track != False and self.overdub_track.clip_slots[0].has_clip:
            self.overdub_track.clip_slots[0].delete_clip()

    def clear_clip(self):
        self.clipStopping = False
        self.remove_clip()

    def copy_clip(self):
        self.send_message("copying clip")
        self.clipSlot.duplicate_clip_to(self.overdub_track.clip_slots[0])
        self.clipStopping = True
        self.clipSlot.clip.stop()
        self.overdub_track.clip_slots[0].clip.fire()
