from .consts import *
import Live


class Clip(object):
    def __init__(self, trackNum, sceneNum, button_num, song, state, parent):
        self.trackNum = trackNum
        self.sceneNum = sceneNum
        self.button_num = button_num
        self.song = song
        self.state = state
        self.__parent = parent
        self.clip_slot = -1
        self.init_listener()
        self.playStatus = -1
        self.muteTimer = Live.Base.Timer(callback=self.on_mute_timer, interval=1, repeat=False)

    def get_clip_slot(self):
        trackNum = self.trackNum
        sceneNum = self.sceneNum
        if self.state.mode == CLIP_LAUNCH_MODE:
            trackNum += self.state.trackOffset
            sceneNum += self.state.sceneOffset
        self.__parent.send_message("track offset: " + str(self.state.trackOffset))
        self.__parent.send_message("scene offset: " + str(self.state.sceneOffset))
        if len(self.song.tracks) >= trackNum:
            track = self.song.tracks[trackNum]
            if len(track.clip_slots) >= sceneNum:
                return track.clip_slots[sceneNum]
        return False

    def init_listener(self):
        self.clip_slot = self.get_clip_slot()
        if self.clip_slot and self.clip_slot.has_clip and not self.clip_slot.clip.color_has_listener(self.update_color):
            self.clip_slot.clip.add_color_listener(self.update_color)
        if not self.clip_slot.has_clip_has_listener(self.has_clip_changed):
            self.clip_slot.add_has_clip_listener(self.has_clip_changed)
        else:
            self.__parent.send_message("listener already added")

    def has_clip_changed(self):
        self.__parent.send_message("in has clip changed")
        self.clip_slot = self.get_clip_slot()
        if self.clip_slot and self.clip_slot.has_clip:
            if not self.clip_slot.clip.playing_status_has_listener(self.on_clip_play_status_changed):
                self.clip_slot.clip.add_playing_status_listener(self.on_clip_play_status_changed)
            if self.clip_slot.clip.is_recording and self.button_num > -1:
                self.__parent.send_sysex(DLCommands.CHANGE_STATE, self.button_num, RECORDING_STATE)
                self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.SLOW_BLINK)
        self.update_color()

    def request_state(self):
        # self.__parent.send_message("trying to update clip state")
        self.update_color()

    def update_color(self):

        self.__parent.send_message("color array update button: " + str(self.button_num) )
        if self.button_num > -1:
            self.clip_slot = self.get_clip_slot()
            self.update_blink()
            if self.clip_slot and self.clip_slot.has_clip:
                carray = self.__parent.get_color_array(self.clip_slot.clip.color)
                if not self.clip_slot.clip.color_has_listener(self.update_color):
                    self.clip_slot.clip.add_color_listener(self.update_color)
                if carray is not None:
                    self.__parent.send_message("color array: " + str(self.button_num))
                    self.__parent.send_sysex(DLCommands.CHANGE_COLOR, self.button_num, carray[0], carray[1], carray[2], carray[3], carray[4], carray[5], True)
                else:
                    self.__parent.send_message("color array none")
            else:
                self.__parent.send_sysex(DLCommands.CHANGE_COLOR, self.button_num, 0, 0, 0, 0, 0, 0, True)
                self.__parent.send_message("clip slot no clip")

    def fire(self, data):
        self.__parent.send_message("triggering clip on track : " + str(self.trackNum + self.state.trackOffset) + " clip number: " + str(self.sceneNum + self.state.sceneOffset))
        self.clip_slot = self.get_clip_slot()

        self.clip_slot.fire()
        self.playStatus = self.clip_slot.clip.is_playing
        if not self.clip_slot.clip.playing_status_has_listener(self.on_clip_play_status_changed):
                self.clip_slot.clip.add_playing_status_listener(self.on_clip_play_status_changed)
        self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.FAST_BLINK)



    def update_blink(self):
        if self.clip_slot.has_clip and not self.clip_slot.clip.is_playing and not self.clip_slot.clip.is_recording:            
            self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.NO_BLINK)
        elif self.clip_slot.has_clip and self.clip_slot.clip.is_recording:
            self.__parent.send_sysex(DLCommands.CHANGE_STATE, self.button_num, RECORDING_STATE)
            self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.FAST_BLINK)
            self.__parent.send_message("cannot play, no clip in slot")
        elif self.clip_slot.has_clip and self.clip_slot.clip.is_playing:
            self.playStatus = self.clip_slot.clip.is_playing
            self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.SLOW_BLINK)
            
    def stop(self, data):
        self.clip_slot = self.get_clip_slot()
        self.__parent.send_message("stopping, data 3 = " + str(data.data[2]))
        if self.clip_slot and self.clip_slot.has_clip and self.clip_slot.clip.is_recording:
            self.clear()
        elif data.data[2] == 1:
            self.clip_slot.stop()
            self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.FAST_BLINK)
        elif self.clip_slot.has_clip:
            if not self.clip_slot.clip.muted_has_listener(self.on_clip_muted):
                self.clip_slot.clip.add_muted_listener(self.on_clip_muted)
                self.clip_slot.clip.muted = True

    def undo(self, data):
        pass

    def clear(self, data = False):
        self.clip_slot = self.get_clip_slot()
        self.__parent.send_message("clearing clip")
        if self.clip_slot and self.clip_slot.has_clip:
            self.clip_slot.delete_clip()

    def on_clip_muted(self):
        self.muteTimer.start()

    def on_mute_timer(self):
        self.clip_slot = self.get_clip_slot()
        clip = self.clip_slot.clip
        clip.muted = False
        clip.remove_muted_listener(self.on_clip_muted)

    def on_clip_play_status_changed(self):
        self.clip_slot = self.get_clip_slot()
        clip = self.clip_slot.clip

        self.__parent.send_message("is playing: " + str(clip.is_playing))

        if self.button_num is not -1:
            self.playStatus = clip.is_playing
            if self.playStatus is True:
                self.__parent.send_message("slow blink on button: " + str(self.button_num))
                self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.SLOW_BLINK)
            elif self.playStatus is False and not clip.is_triggered:
                self.__parent.send_message("no blink")
                self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.NO_BLINK)
            elif self.playStatus is False and clip.is_triggered:
                self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.FAST_BLINK)
        # self.update_color()

    def remove(self):
        if self.clip_slot and self.clip_slot.has_clip and self.clip_slot.clip.color_has_listener(self.update_color):
            self.clip_slot.clip.remove_color_listener(self.update_color)
        if self.clip_slot.has_clip_has_listener(self.update_color):
            self.clip_slot.remove_has_clip_listener(self.update_color)
        self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.NO_BLINK)
