from .cltrack import ClTrack
from .consts import *
import Live

class ClMidiTrack(ClTrack):

    def __init__(self, parent, track, trackNum, song, state, action_handler):
        super(ClMidiTrack, self).__init__(parent, track, trackNum, song , state, action_handler)
        self.__parent = parent
        self.trackStore = []
        self.state = state
        self.prevNotes = list()

    ### SCENARIOS ###
    # 1. No clip in memory, record starts recording and a new clip is made and selected
    # 2. Clip is already in memory and playing, meaning overdubbing will begin
    # 3. Clip is already in memory and stopped, meaning clip will play
    # 4. Clip is already in memory and recording, meaning clip will play
    #################
    def record(self, quantized):
        self.__parent.send_message("in record pressed, track num " + str(self.track.name))
        if self.clipSlot == -1 and self.track.arm:
            self.__parent.send_message("starting recording in new slot")
            # Scenario # 1
            self.get_new_clip_slot(False)
            self.fire_clip(quantized)
        elif self.clipSlot != -1 and self.clipSlot.has_clip and self.track.arm:
            if not self.clipSlot.clip.is_playing and not self.clipSlot.clip.is_recording:
                self.__parent.send_message("playing clip from stopped state")
                # Scenario 3
                self.fire_clip(quantized)
            elif self.clipSlot.clip.is_recording and not self.clipSlot.clip.is_overdubbing:
                # Scenario 4
                self.__parent.send_message("ending recording")
                if self.state.clip_quantization != Live.Song.Quantization.q_no_q:
                    self.clipSlot.clip.quantize(self.state.clip_quantization, 1)
                self.fire_clip(quantized)
            elif self.clipSlot.clip.is_overdubbing:
                self.endOverdub()
        elif self.clipSlot != -1 and not self.clipSlot.has_clip  and self.track.arm:
            self.__parent.send_message("recording new clip")
            # Scenario 5
            self.fire_clip(quantized)

    def stop(self, quantized):
        if self.track.arm:
            super(ClMidiTrack, self).stop(quantized)
            if self.clipSlot != -1 and self.clipSlot.has_clip and self.clipSlot.clip.is_overdubbing:
                self.endOverdub()

    def undo(self, fadeTime = 0):
        self.undoOverdub()

    def overdub(self):
        if self.track.arm:
            self.__parent.send_message("overdubbing")
            self.update_state(OVERDUB_STATE)
            self.__parent.session_record(True, self.trackNum)
            if self.clipSlot != -1 and self.clipSlot.has_clip != -1:
                self.clipSlot.clip.select_all_notes()
                self.prevNotes.append(self.clipSlot.clip.get_selected_notes())
            self.song.session_record = 1

    def undoOverdub(self):
        if len(self.prevNotes) > 0 and self.clipSlot != -1 and self.clipSlot.has_clip:
            self.__parent.send_message("removing overdub")
            self.clipSlot.clip.select_all_notes()
            self.clipSlot.clip.replace_selected_notes(self.prevNotes[-1])
            del self.prevNotes[-1]

    def endOverdub(self):
        self.__parent.send_message("ending overdubbing")
        self.song.session_record = 0
        self.update_state(PLAYING_STATE)
        self.__parent.session_record(False, self.trackNum)

    def new_clip(self):
        self.get_new_clip_slot(False)

    def get_new_clip_slot(self, new_scene = False):
        if self.lastState == OVERDUB_STATE:
            self.endOverdub()
        super(ClMidiTrack, self).get_new_clip_slot(new_scene)

    def remove_clip_slot(self):
        if self.lastState == OVERDUB_STATE:
            self.endOverdub()
        super(ClMidiTrack, self).remove_clip_slot()

    def trigger_record(self, quantized = True):
        self.overdub()

