NUM_TRACKS = 3
NUM_CONTROLS = 6

MASTER_CONTROL = 0
STOP_CONTROL = 1
UNDO_CONTROL = 2
CLEAR_CONTROL = 3
PLAY_CONTROL = 4
RECORD_CONTROL = 5

REC_DELAY_TOGGLED = 16
FIXED_LENGTH_REC_TOGGLED = 17
DATALOOPER_VENDOR = 30
THIRD_PARTY_VENDOR = 31

class DLCommands:
    ABLETON_CONNECTED = 0
    CHANGE_STATE = 1
    REQUEST_MIDI = 2
    ON_BEAT = 3
    CHANGE_COLOR = 4
    BLINK = 5
    SONG_STOPPED = 11
    CHANGE_MODE = 12
    KEEP_ALIVE = 13
    RESTORE_COLOR = 14

DATALOOPER_KEY = 'DL#'
CLIPLOOPER_KEY = 'CL#'
GROUP_KEY = 'G#'
GROUP_MASTER_KEY = 'M#'
OVERDUB_KEY = 'OVR'
LED_KEY = 'LED'
TRACK_KEYS = [DATALOOPER_KEY, CLIPLOOPER_KEY]
ALT_KEYS = [GROUP_KEY, GROUP_MASTER_KEY,OVERDUB_KEY,LED_KEY]
key_map = {
        GROUP_KEY : "set_group",
        GROUP_MASTER_KEY: "set_group_master",
        OVERDUB_KEY: "set_overdub",
        LED_KEY: "set_led",
        }
NOTE_OFF_STATUS = 128
NOTE_ON_STATUS = 144
CC_STATUS = 176

# QUANTIZE VALUES
GLOBAL = 0.0
NO_QUANTIZATION = 1.0
EIGHT_BARS = 2.0
FOUR_BARS = 3.0
TWO_BARS = 4.0
ONE_BAR = 5.0
HALF_NOTE = 6.0
HALF_TRIPLET = 7.0
QUARTER_NOTE = 8.0
QUARTER_TRIPLET = 9.0
EIGHTH_NOTE = 10.0
EIGHTH_TRIPLET = 11.0
SIXTEENTH_NOTE = 12.0
SIXTEENTH_NOTE_TRIPLET = 13.0
THIRTY_SECOND_NOTE = 14.0

# LOOPER PARAMS
DEVICE_ON = 0
STATE = 1

STOP_STATE = 0
RECORDING_STATE = 1
PLAYING_STATE = 2
OVERDUB_STATE = 3
CLEAR_STATE = 4
OFF_STATE = 5
NEW_SESSION_STATE = 6
CONFIG_STATE = 7
DISARMED_STATE = 8

FEEDBACK = 2
REVERSE = 3
MONITOR = 4
SPEED = 5
QUANTIZATION = 6
SONG_CONTROL = 7

NO_TEMPO_CONTROL = 0
FOLLOW_SONG_TEMPO = 1
SET_AND_FOLLOW_SONG_TEMPO = 2

TEMPO_CONTROL = 8

ANY_VALUE = -1

LOOPER_TRACK = 0
CLIP_TRACK = 1
BOTH_TRACK_TYPES = 2

LOOPER_MODE = 0
NEW_SESSION_MODE = 1
CLIP_LAUNCH_MODE = 2

OFF = 0
ON = 1
TOGGLE = 2

DELETE_CLIP = 0
NEW_SCENE_KEEP_PLAYING = 1
NEW_SCENE_STOP = 2
STOP_CLIP = 3

LOOPER_ACTIONS = {
    0 : 'record_quantized',
    1 : 'record_immediately',
    14 : 'exclusive_record_quantized',
    15 : 'exclusive_record_immediately',
    2 : 'stop_quantized',
    3 : 'stop_immediately',
    4 : 'undo',
    5 : 'clear',
    7 : 'none',
    12 : 'overdub',
    8 : 'toggle_mute',
    9 : 'remove_clip_slot',
    10 : 'record_ignoring_state',
    11 : 'play',
    13 : 'stop_play_toggle',
    16 : 'clear_ignoring_rec_arm'
}

class PressState:
    PRESS = 0
    RELEASE = 1
    DOUBLE_TAP = 2
    DOUBLE_TAP_RELEASE = 3
    LONG_PRESS = 4
    LONG_PRESS_RELEASE = 5


QUANTIZATION_BEATS = {
    "q_2_bars" : 16,
    "q_4_bars" : 32,
    "q_8_bars" : 64,
    "q_bar" : 4,
    "q_eight" : .5,
    "q_eight_triplet" : .666,
    "q_half" : 2,
    "q_half_triplet" : 1.5,
    "q_no_q" : 0,
    "q_quarter" : 1,
    "q_quarter_triplet" : 1.333,
    "q_sixteenth" : .25,
    "q_sixteenth_triplet" : .333,
    "q_thirtywoth" : .125
}

class DLActions:
    action_map = {
        0: "metronome_control",
        1: "looper_control",
        2: "tap_tempo",
        3: "clip_control",
        4: "toggle_stop_start",
        5: "mute_control",
        6: "transport_control",
        7: "scene_control",
        8: "change_mode",
        9: "change_instance",
        10: "move_session_highlight",
        11: "request_state",
        12: "request_midi_map_rebuild",
        13: "set_group_masters",
        14: "start_recording",
        15: "solo_control",
        16: "set_record_delay",
        17: "set_record_length",
        18: "smart_clear",
        19: "load_song",
        20: "launch_song_part"
        }
    METRONOME_CONTROL = 0
    LOOPER_CONTROL = 1
    TAP_TEMPO = 2
    CLIP_CONTROL = 3
    TOGGLE_STOP_START = 4
    MUTE_CONTROL = 5
    TRANSPORT_CONTROL = 6
    SCENE_CONTROL = 7
    CHANGE_MODE = 8
    CHANGE_INSTANCE = 9
    MOVE_SESSION_HIGHLIGHT = 10
    REQUEST_STATE = 11
    REQUEST_MIDI_MAP_REBUILD = 12
    SET_GROUP_MASTERS = 13
    START_RECORDING = 14
    SET_RECORD_DELAY = 15
    SET_RECORD_LENGTH = 16

class LooperActions:
    RECORD_QUANTIZED = 0
    RECORD_IMMEDIATELY = 1
    EXCLUSIVE_RECORD_QUANTIZED = 14
    EXCLUSIVE_RECORD_IMMEDIATELY = 15
    NONE = 7
    STOP_QUANTIZED = 2
    STOP_IMMEDIATELY = 3
    UNDO = 4
    CLEAR = 5
    OVERDUB = 12
    TOGGLE_MUTE = 8
    REMOVE_CLIP_SLOT = 9
    RECORD_IGNORING_STATE = 10
    PLAY = 11
    STOP_PLAY_TOGGLE = 13
    RECORD_ACTIONS = {RECORD_QUANTIZED, RECORD_IMMEDIATELY, EXCLUSIVE_RECORD_IMMEDIATELY, EXCLUSIVE_RECORD_QUANTIZED}
    FADEABLE_ACTIONS = {CLEAR, STOP_QUANTIZED}


ALL_LOOPERS = -1


CLIP_ACTIONS = {
    0 : 'fire',
    1 : 'stop',
    2 : 'clear',
    3 : 'undo'
}

class BlinkTypes():
    NO_BLINK = 0
    SLOW_BLINK = 1
    FAST_BLINK = 2

class LiveOSCBridge():
    script = None

    @staticmethod
    def link_bridge(script):
        if LiveOSCBridge.script is None:
            LiveOSCBridge.script = script