from __future__ import with_statement
import Live
from _Framework.ControlSurface import ControlSurface
from .consts import *
from .trackhandler import TrackHandler
from .sysex import Sysex
from .actions import Actions
from collections import defaultdict
from .state import State
import json
import os
from _Framework.MixerComponent import MixerComponent # Class encompassing several channel strips to form a mixer
from _Framework.EncoderElement import *
from _Framework.ButtonElement import ButtonElement

class DataLooper(ControlSurface):

    def __init__(self, c_instance):
        super(DataLooper, self).__init__(c_instance)
        self.__c_instance = c_instance

        # shows in Ableton footer
        self.show_message("Powered by DATA Looper")

        # listens to time changes
        self.song().add_current_song_time_listener(self.on_time_change)

        # listens to track add/remove
        self.song().add_tracks_listener(self.on_track_added_or_removed)

        self.song().add_signature_numerator_listener(self.on_signature_numerator_change)

        self.tracks = defaultdict(list)
        # self.keep_alive_timer = Live.Base.Timer(callback=self.send_keep_alive, interval=1000, repeat=True)
        self.state = State(self.song())
        self.alt_loopers = []
        # creates obj to handle actions
        self.__action_handler = Actions(self, self.tracks, self.song(), self.state)

        # creates obj to handle tracks
        self.__track_handler = TrackHandler(self, self.song(), self.tracks, self.state, self.__action_handler)

        self.song().add_is_playing_listener(self.on_is_playing)
        self.__track_handler.scan_tracks(False)
        # initializes base obj
        self.live = Live.Application.get_application()
        # self.keep_alive_timer.start()
        self.song().add_metronome_listener(self.on_metro_change)
        self.send_sysex(0x00, 0x01)
        


        # with self.component_guard():
        #     global _map_modes
        #     _map_modes = Live.MidiMap.MapMode
        #     # mixer
        #     global mixer
        #     num_tracks = 128
        #     num_returns = 24
        #     self.mixer = MixerComponent(num_tracks, num_returns)
        #     self.map_tracks()

    # def map_tracks(self):
    #     # activate mode
    #     for i in range(0, len(self.song().tracks)):
    #         self.mixer.channel_strip(i).set_volume_control(EncoderElement(MIDI_CC_TYPE, 0, i, _map_modes.absolute))
    #         self.mixer.channel_strip(i).set_pan_control(EncoderElement(MIDI_CC_TYPE, 1, i, _map_modes.absolute))
    #         send_controls = [EncoderElement(MIDI_CC_TYPE, 2, i, _map_modes.absolute), EncoderElement(MIDI_CC_TYPE, 3, i, _map_modes.absolute),
    #                          EncoderElement(MIDI_CC_TYPE, 4, i, _map_modes.absolute) ];
    #         self.mixer.channel_strip(i).set_send_controls(send_controls)
    #         self.mixer.channel_strip(i).set_send_controls(send_controls)
    #         self.mixer.channel_strip(i).set_solo_button(ButtonElement(True, MIDI_CC_TYPE, 6, i))

    def on_is_playing(self):
        if not self.song().is_playing:
            self.send_sysex(DLCommands.SONG_STOPPED);

    def send_keep_alive(self):
        self.send_sysex(DLCommands.KEEP_ALIVE)

    def on_metro_change(self):
        if not self.state.ignoreMetroCallback:
            self.state.tap_tempo_counter = 0
            self.state.ignoreMetroCallback = False

    def send_message(self, m):
        self.log_message("Instance: " + str(id(self)) + " " + str(m))
        pass

    def set_bpm(self, bpm):
        self.song().tempo = bpm

    def on_track_added_or_removed(self):
        self.send_message("on track change")
        self.__track_handler.scan_tracks()

    def on_track_name_changed(self):
        self.send_message("on track name change")
        self.__track_handler.scan_tracks()

    def on_time_change(self):
        time = self.song().get_current_beats_song_time()

        if time.beats != self.state.curBeats:
            self.state.curBeats = time.beats
            self.send_sysex(DLCommands.ON_BEAT, self.state.curBeats)

    def on_signature_numerator_change(self):
        self.send_message("numerator changed")
        self.send_midi((0xF0, 0x1E, 0x01, self.song().signature_numerator, 0xF7))

    def connect(self):
        self.log_message("looper connecting")
        self.send_sysex(0x00, 0x01)
        

    def disconnect(self):
        self.log_message("looper disconnecting")
        self.send_sysex(0x00, 0x00)

        super(DataLooper, self).disconnect()

    def send_midi(self, midi_event_bytes):
        self.__c_instance.send_midi(midi_event_bytes)

    def send_sysex(self, *data): 
        # check for nested tuple
        for d in data:
            if type(d) is tuple:
                data = d
                break
        # self.send_message("sending sysex: " + str(datastring))
        sysex = (0xF0, 0x1E) + data + (0xF7,)
        self.send_midi(sysex)

    def send_program_change(self, program):
        self.send_message("sending program change:" + str(program))
        looper_status_sysex = (192, program)
        self.send_midi(looper_status_sysex)

    def receive_midi(self, midi_bytes):
        # super(DataLooper,self).receive_midi(midi_bytes)
        if len(midi_bytes) != 3:
            self.handle_sysex(midi_bytes)
        else:
            self.send_message("non sysex midi received")
    
    def get_connection_object(self):
        return LiveOSCBridge()
        
    def add_looper_state_listener(self, listener):
        self.__track_handler.add_looper_state_listener(listener)

    def receive_midi_cc(self, cc_no, cc_value, channel):
        self.log_message(str(cc_no))

    def handle_sysex(self, midi_bytes):
        self.send_message("sysex vendor " + str(midi_bytes[1]))
        if midi_bytes[1] == DATALOOPER_VENDOR or midi_bytes[1] == THIRD_PARTY_VENDOR:
            sysex = Sysex(midi_bytes)
            self.send_message("midi action byte:")
            self.send_message(sysex.action)
            self.send_message(self.get_method(sysex.action))
            getattr(self.__action_handler, self.get_method(sysex.action))(sysex)
            
            # This would send any incoming sysex commands to the other Datalooper instances. I don't remember why I did this, and it was causing issues, so it's commented out. 
            # Generally, each instance should be unique and rely on listeners to change state rather than explicit commands from another pedal.
            #if sysex.data[1] == LooperActions.CLEAR and not sysex.is_passed_from_another_looper:
                # May be a problem with duplicating fade
             #   self.send_command_to_other_loopers(sysex)

    def send_command_to_other_loopers(self, sysex):
        sysex.is_passed_from_another_looper = True
        self.send_message("sending command to alt loopers: " + str(len(self.alt_loopers)))
        for control in self.live.control_surfaces:
            if isinstance(control, DataLooper) and control != self:
                control.execute_method(sysex)

    def set_clip_quantization(self, clip_quantization):
        self.send_message("linkage worked, sending value:" + str(clip_quantization))
        self.state.set_clip_quantization(clip_quantization)

    def execute_method(self, sysex):
        getattr(self.__action_handler, self.get_method(sysex.action))(sysex)

    def refresh_state(self):
        """Live -> Script
        Send out MIDI to completely update the attached MIDI controller.
        Will be called when requested by the user, after for example having reconnected
        the MIDI cables...
        """
        self.send_sysex(0x00, 0x01)
        # self.send_message("refreshing state")
        super(DataLooper, self).refresh_state()

    @staticmethod
    def get_method(argument):
        return DLActions.action_map.get(argument, "Invalid Action")
