from .consts import *
import Live
from .track import Track
from time import time

class DlTrack(Track):
    """ Class handling looper track from DataLooper """

    def __init__(self, parent, track, device, trackNum, song, state, action_handler):
        super(DlTrack, self).__init__(parent, track, trackNum, song, state, action_handler)
        # self.updateReq = False
        # self.name_timer = Live.Base.Timer(callback=self.change_name, interval=1, repeat=False)
        self.tempo_control = -1
        self.device = device
        self.state = device.parameters[STATE]
        self.rectime = 0
        self.ignore_stop = False
        self.req_record = True
        self.__parent = parent
        self.tempState = ""
        self.undoState = False
        self.isDuplicate = False
        self.restore_stop_state_after_undo = False
        self.send_message("initializing new track num: " + str(self.trackNum))
        # self.ignore_name_change = False
        self.device.parameters[TEMPO_CONTROL].add_value_listener(self.on_tempo_control_change)
        # self.device.add_name_listener(self.on_name_change)
        if self.track.can_be_armed:
            self.track.add_arm_listener(self.set_arm)

        if not self.state.value_has_listener(self._on_looper_param_changed):
            self.state.add_value_listener(self._on_looper_param_changed)

    def set_duplicate(self, isDuplicate):
        self.isDuplicate = isDuplicate

    def set_arm(self):
        self.update_state(CLEAR_STATE)

    def _on_looper_param_changed(self):
        if self.lastState == CLEAR_STATE and self.state.value == STOP_STATE:
            return
        elif self.lastState != self.state.value:
            self.send_message("Looper param changed. Last State: " + str(self.lastState) + " New State: " + str(self.state.value))
            self.send_message("device name:" + str(self.device.name))
            self.update_state(int(self.state.value))

    def send_message(self, message):
        self.__parent.send_message(message)

    def on_tempo_control_change(self):
        if not self.global_state.ignore_tempo_control:
            # self.send_message("changing mode via listener")
            if self.device.parameters[TEMPO_CONTROL].value == 0:
                self.__parent.send_sysex(DLCommands.CHANGE_MODE, 1)
            else:
                self.__parent.send_sysex(DLCommands.CHANGE_MODE, 0)

    def stop_play_toggle(self):
        if self.lastState == STOP_STATE:
            self.request_control(MASTER_CONTROL)
        elif self.lastState == PLAYING_STATE or self.lastState == CLEAR_STATE:
            self.request_control(STOP_CONTROL)
        elif self.lastState == RECORDING_STATE:
            self.request_control(CLEAR_CONTROL)

    def request_control(self, controlNum):
        if not self.isDuplicate:
            self.send_message("Requesting control: ")
            # self.updateReq = True
            self.__parent.send_sysex(DLCommands.REQUEST_MIDI, self.trackNum, (self.trackNum * NUM_CONTROLS) + controlNum)

    def record(self, quantized):
        #self.__parent.send_message("Looper " + str(self.trackNum) + " state: " + str(self.device.parameters[STATE].value) + " rec pressed")
        if self.rectime == 0 or (time() - self.rectime) > .5:
            if not quantized and self.song.is_playing and self.lastState == CLEAR_STATE:
                self.send_message("updating to record state")
                # self.updateReq = True
                self.state.value = RECORDING_STATE
                self.rectime = time()
            elif not quantized and self.song.is_playing and self.lastState == RECORDING_STATE:
                self.state.value = STOP_STATE
                self.calculateBPM(time() - self.rectime)
            elif not quantized and self.lastState == STOP_STATE:
                # self.updateReq = True
                self.state.value = PLAYING_STATE
            else:
                self.trigger_record()

    def trigger_record(self):
        self.send_message("rectime: " + str(time() - self.rectime))
        self.rectime = time()
        self.request_control(MASTER_CONTROL)
        if self.button_num != -1:
            self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.FAST_BLINK)

    #Triggered on all tracks from actions.py. Ends recording on other loopers and records on the currently selected one
    def exclusive_record(self, quantized = True, looperNum = -1):
        self.send_message("exclusive record looperNum: " + str(looperNum) + " Track Num: " + str(self.trackNum))
        if looperNum == self.trackNum:
            self.record(quantized)
        elif self.lastState == RECORDING_STATE:
            self.trigger_record()

    def record_ignoring_state(self):
        self.request_control(RECORD_CONTROL)

    def play(self, quantized=True):
        self.send_message("attempting play, last state: " + str(self.lastState) + " track is: " + str(self.track.name))
        if self.lastState == STOP_STATE:
            
            if quantized:
                self.request_control(MASTER_CONTROL)
            else:
                self.state.value = PLAYING_STATE
            self.update_state(PLAYING_STATE)

    def stop_quantized(self, fadeTime = 0):
        if self.lastState == RECORDING_STATE:
            self.clear_immediately()
        elif (self.lastState != STOP_STATE and self.lastState != CLEAR_STATE) or not self.song.is_playing:
            super(DlTrack, self).stop_quantized(fadeTime)

    def stop_immediately(self, fadeTime = 0):
        self.__parent.send_message("stop immediately")
        self.stop(False)

    def stop(self, quantized):
        self.__parent.send_message(
            "Looper " + str(self.track.name) + " state: " + str(self.lastState) + " stop pressed")
        if self.lastState == RECORDING_STATE:
            self.request_control(CLEAR_CONTROL)
            self.__parent.send_message("last state was clear, clearing loop from stop")
            self.ignore_stop = True
            self.update_state(CLEAR_STATE)
        if self.lastState != CLEAR_STATE:
            if quantized or not self.song.is_playing:
                self.request_control(STOP_CONTROL)
                if self.song.is_playing and self.button_num != -1:
                    self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.FAST_BLINK)
            else:
                self.send_message("entering stop state")
                # self.updateReq = True
                self.update_state(STOP_STATE)
                self.state.value = STOP_STATE

        # SEND OUT CONTROL NO MATTER WHAT IF SONG IS NOT PLAYING, FOR MAPPING PARAMS
        elif not self.song.is_playing:
            self.request_control(STOP_CONTROL)
        

    def toggle_playback(self):
        if self.lastState == STOP_STATE:
            self.request_control(MASTER_CONTROL)
        elif self.lastState == PLAYING_STATE:
            self.request_control(STOP_CONTROL)

    def undo(self, fadeTime = 0):
        if (self.lastState != CLEAR_STATE and self.lastState != RECORDING_STATE) or not self.song.is_playing:
            self.send_message("requesting undo command")
            self.request_control(UNDO_CONTROL)
            self.undoState = not self.undoState
        elif self.restore_stop_state_after_undo:
            self.restore_stop_state_after_undo = False
            self.request_control(UNDO_CONTROL)
            self.update_state(STOP_STATE)
        else:
            self.clear_immediately()
            # if self.undoState:
            #     self.__parent.send_sysex(DLCommands.CHANGE_COLOR, self.undo_button, 127, 127, 127, 127, 127, 127)
            # else:
            #     self.__parent.send_sysex(DLCommands.CHANGE_COLOR, self.undo_button, 0, 0, 0, 0, 0, 0)



        # self.__parent.send_message(
        #     "Looper " + str(self.trackNum) + " state: " + str(self.device.parameters[1].value) + " undo pressed")

    def clear_immediately(self):
        super(DlTrack, self).clear_immediately()
        if self.lastState == OVERDUB_STATE:
            self.state.value = STOP_STATE
        elif self.lastState != CLEAR_STATE:
            self.restore_stop_state_after_undo = True    

        if self.track.devices and self.track.devices[0].can_have_chains and hasattr(self.track.devices[0],'variation_count') and self.track.devices[0].variation_count and self.track.devices[0].name == "Datalooper FX":
            self.track.devices[0].selected_variation_index = 0  
            self.track.devices[0].recall_selected_variation() 
        self.request_control(CLEAR_CONTROL)
        # self.__parent.send_message(
        #     "Looper " + str(self.trackNum) + " state: " + str(self.device.parameters[1].value) + " clear pressed")
        self.update_state(CLEAR_STATE)

    def calculateBPM(self, loop_length):
        loop_length_in_minutes = loop_length / 60
        i = 1
        bpms = []
        closest_to_tempo = 120

        while i <= 64:
            i *= 2
            bpm = i / loop_length_in_minutes
            if 140 > bpm > 50:
                bpms.append(bpm)

        bpm = min(bpms, key=lambda x: abs(x - closest_to_tempo))
        self.send_message("bpm: " + str(bpm))
        self.global_state.queued = self
        self.rectime = 0
        self.global_state.bpm = bpm
        self.action_handler.jump_to_next_bar()

    def change_mode(self):
        # self.send_message("changing mode")
        if self.device:
            if self.global_state.mode == LOOPER_MODE:
                self.device.parameters[TEMPO_CONTROL].value = SET_AND_FOLLOW_SONG_TEMPO
            elif self.global_state.mode == NEW_SESSION_MODE:
                self.tempo_control = self.device.parameters[TEMPO_CONTROL].value
                self.device.parameters[TEMPO_CONTROL].value = NO_TEMPO_CONTROL

    def remove_track(self):
        if self.track in self.song.tracks:
            if self.track.can_be_armed and self.track.arm_has_listener(self.set_arm):
                self.track.remove_arm_listener(self.set_arm)
            if self.state is not None and self.state.value_has_listener(self._on_looper_param_changed):
                self.state.remove_value_listener(self._on_looper_param_changed)
        else:
            self.send_message("removing track: " + str(self.trackNum))
            self.update_state(OFF_STATE)


    def update_state(self, state):
        if state == CLEAR_STATE:
            self.undoState = False

        super(DlTrack, self).update_state(state)

    # def change_name(self):
    #     self.send_message("changing name on dl#" + str(self.trackNum) + " to: " + str(self.lastState)  )
    #     self.ignore_name_change = True
    #     self.device.name = str(self.tempState)
    #
    # def on_name_change(self):
    #     if not self.ignore_name_change and str(self.device.name) == str(CLEAR_STATE):
    #         self.send_message("updating state based on device name")
    #         self.update_state(int(self.device.name))
    #     self.ignore_name_change = False

    def clear(self, fadeTime = 0):
        if self.lastState == CLEAR_STATE or self.lastState == RECORDING_STATE or self.lastState == STOP_STATE:
            self.clear_immediately()
        else:
            super(DlTrack, self).clear(fadeTime)