from .consts import *
from .dltrack import DlTrack
from .cltrack import ClTrack
from .track import Track

class Mute:
    def __init__(self, parent, button_num, mute_type, mute_what, song, tracks, trackNum = -1):
        self.tracks = tracks
        self.__parent = parent
        self.song = song
        self.mute_type = mute_type
        self.button_num = button_num
        self.mute_what = mute_what
        self.is_muted = False
        self.trackNum = trackNum
        self.looper_track = -1
        if self.trackNum >= 0:
            looper_tracks = self.tracks.get(str(self.trackNum))
            if looper_tracks:
                for looper_track in looper_tracks:
                    self.looper_track = looper_track
                    looper_track.track.add_mute_listener(self.on_mute_change)
                    return
        #mute what:
        # 0 = tracks playing clips
        # 1 = all tracks
        # 2 = all dl# tracks
        # 3 = all cl# tracks
        # 4 = all cl# and dl# tracks

        #mute type:
        # 0 = mute
        # 1 = unmute
        # 2 = toggle

    def on_mute_change(self):
        if self.looper_track != -1 and self.button_num >= 0:
            self.__parent.send_message("sending mute state on button number: " + str(self.button_num) + " track name: " + str(self.looper_track.track.name))
            if self.looper_track.track.mute:
                self.is_muted = True
                self.__parent.send_sysex(DLCommands.CHANGE_STATE, self.button_num, OFF_STATE)
            else:
                self.__parent.send_sysex(DLCommands.CHANGE_STATE, self.button_num, CLEAR_STATE)
                self.is_muted = False




    def request_state(self):
        self.on_mute_change()

    def execute_mute(self):
        self.__parent.send_message("executing mute on buttonNum: " + str(self.button_num) + " mute_type: " + str(self.mute_type) + " mute what: " + str(self.mute_what))
        # Get the function from switcher dictionary

        if self.mute_type == 2:
            self.is_muted = not self.is_muted
        else:
            self.is_muted = self.mute_type  

        getattr(self, self.get_method(self.mute_what))(self.is_muted)

        self.request_state()

    def mute_tracks_playing_clips(self, shouldMute):
        for track in self.song.tracks:
            if track.playing_slot_index > -1:
                track.mute = shouldMute

    def mute_all_tracks(self, shouldMute):
        for track in self.song.tracks:
            track.mute = shouldMute

    def mute_all_dl_tracks(self, shouldMute):
        for tracks in self.tracks.values():
            for track in tracks:
                if isinstance(track, DlTrack):
                    track.track.mute = shouldMute

    def mute_all_cl_tracks(self, shouldMute):
        for tracks in self.tracks.values():
            for track in tracks:
                if isinstance(track, ClTrack):
                    track.track.mute = shouldMute

    def mute_all_looper_tracks(self, shouldMute):
        for tracks in self.tracks.values():
            for track in tracks:
                track.track.mute = shouldMute

    def mute_single_track(self, shouldMute):
        if self.trackNum >= 0:
            self.__parent.send_message("muting")
            tracks = self.tracks.get(str(self.trackNum))
            for track in tracks:
                track.track.mute = shouldMute

    @staticmethod
    def get_method(argument):
        action_map = {
            0: "mute_tracks_playing_clips",
            1: "mute_all_tracks",
            2: "mute_all_dl_tracks",
            3: "mute_all_cl_tracks",
            4: "mute_all_looper_tracks",
            5: "mute_single_track"
        }
        return action_map.get(argument, "Invalid Action")

    def remove(self):
        pass