from .consts import *
import Live

class Scene:
    def __init__(self, sceneNum, button_num, song, state, parent):
        self.sceneNum = sceneNum
        self.button_num = button_num
        self.song = song
        self.state = state
        self.__parent = parent
        self.init_listeners()
        self.scene_timer = -1
        self.queued_scene = -1
        self.last_fired = -1

    def init_listeners(self):
        if self.sceneNum < len(self.song.scenes):
            if not self.song.scenes[self.sceneNum].color_has_listener(self.on_color_change):
                self.song.scenes[self.sceneNum].add_color_listener(self.on_color_change)
            else:
                self.__parent.send_message("color listener already added")

            if not self.song.scenes[self.sceneNum].is_triggered_has_listener(self.on_scene_triggered):
                self.song.scenes[self.sceneNum].add_is_triggered_listener(self.on_scene_triggered)
            else:
                self.__parent.send_message("color listener already added")

            self.update_color()

    def next_scene(self):
        if self.last_fired == -1:
            self.song.scenes[0].fire()
            self.last_fired = 0
        else:
            self.last_fired += 1
            self.song.scenes[self.last_fired].fire()

    def prev_scene(self):
        pass

    def on_scene_triggered(self):
        self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.SLOW_BLINK)

    def stop_blink(self):
        self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.NO_BLINK)

    def request_state(self):
        if self.sceneNum < len(self.song.scenes) and self.button_num > -1:
            self.update_color()

    def on_color_change(self):
        self.update_color()

    def update_color(self):
        carray = self.__parent.get_color_array(self.song.scenes[self.sceneNum].color)
        self.__parent.send_sysex(DLCommands.CHANGE_COLOR, self.button_num, carray[0], carray[1], carray[2], carray[3],carray[4], carray[5], True)

    def remove(self):
        if self.song.scenes[self.sceneNum] and self.song.scenes[self.sceneNum].color_has_listener(self.on_color_change):
            self.song.scenes[self.sceneNum].remove_color_listener(self.on_color_change)

    def fire_scene(self, should_queue):
        clip_repeat_lengths = []
        self.__parent.send_message("attempting to fire scene: " + str(self.sceneNum) + "should queue? : " + str(should_queue) )
        if should_queue:
            for track in self.song.tracks:
                if track.playing_slot_index > -1:
                    clip = track.clip_slots[track.playing_slot_index].clip
                    cur_pos = clip.playing_position
                    loop_pos = clip.loop_end
                    clip_repeat_lengths.append(loop_pos - cur_pos)
            if len(clip_repeat_lengths):
                beats = self.song.current_song_time
                self.__parent.send_message("beats:" + str(beats))
                trigger_seconds = self.beats_to_seconds(beats % QUANTIZATION_BEATS.get(str(self.song.clip_trigger_quantization)))
                seconds_to_next_repeat = self.beats_to_seconds(max(clip_repeat_lengths))
                if seconds_to_next_repeat <= trigger_seconds:
                    self.song.scenes[self.sceneNum].fire()
                else:
                    self.queued_scene = self.sceneNum
                    ms_to_fire = int((seconds_to_next_repeat - (trigger_seconds) )* 1000)
                    self.__parent.send_message("seconds to fire:" + str(ms_to_fire) + " trigger seconds: " + str(trigger_seconds))
                    self.scene_timer = Live.Base.Timer(callback=self.fire, interval=ms_to_fire, repeat=False)
                    self.scene_timer.start()
                self.__parent.send_message("max seconds till clip repeats: " + str(max(clip_repeat_lengths)))
            else:
                self.queued_scene = self.sceneNum
                self.fire()
        else:
            self.queued_scene = self.sceneNum
            self.fire()

    def fire(self):
        self.__parent.send_message("firing scene: " + str(self.queued_scene))
        if self.queued_scene < 125:
            self.song.scenes[self.queued_scene].fire()
            for scene in self.__parent.scenes:
                scene.stop_blink()
        elif self.queued_scene == 125:
            self.next_scene()

    def beats_to_seconds(self, beats):
        return (beats / self.song.tempo) * 60