from .consts import *
from .dltrack import DlTrack
from .cltrack import ClTrack
from .track import Track

class Solo:
    def __init__(self, parent, song, tracks):
        self.looper_tracks = tracks
        self.__parent = parent
        self.song = song
        self.tracks = song.tracks
        self.muted_tracks = []
        self.linked_buttons = {}
        self.group_tracks = []
        self.song.add_tracks_listener(self.on_tracks_changed)
        self.on_tracks_changed()

    def on_tracks_changed(self):
        self.group_tracks = []
        self.tracks = self.song.tracks
        for track in self.tracks:
            if hasattr(track, 'group_track') and track.group_track and track.group_track not in self.group_tracks:
                self.__parent.send_message("adding group track to solo exclude: " + str(track.group_track.name))
                self.group_tracks.append(track.group_track)
            if "NO-SOLO" in track.name:
                self.__parent.send_message("adding no solo track to solo exclude: " + str(track.name))
                self.group_tracks.append(track)

    def request_state(self, track_number):
        pass
        # if track_number in self.linked_buttons:
        #     button_number = self.linked_buttons[track_number]
        #     self.__parent.send_message("sending solo state")
        #     requested_tracks = self.looper_tracks.get(str(track_number))
        #     for requested_track in requested_tracks:
        #         if requested_track in self.soloed_tracks:
        #             self.__parent.send_sysex(DLCommands.CHANGE_STATE, button_number, CLEAR_STATE)
        #             return
        #         else:
        #             self.__parent.send_sysex(DLCommands.CHANGE_STATE, button_number, OFF_STATE)

    def execute_solo(self, track_num, mute_all, exclusive_solo):

        self.__parent.send_message("executing solo on track number: " + str(track_num))

        #unsoloing all tracks if track num -1
        if track_num is -1:
            for track in self.tracks:
                track.solo = False
            return
        #gets all tracks corresponding to #, (DL#1 Keys, DL#1 Flute)
        requested_solo_tracks = self.looper_tracks.get(str(track_num))

        #unsolos tracks
        for requested_solo_track in requested_solo_tracks:
            if requested_solo_track.track.solo:
                self.__parent.send_message("unsoloing track: " + str(requested_solo_track.track.name) + " solo? " + str(requested_solo_track.track.solo))
                self.unsolo_track(requested_solo_tracks)
                return 

        #if the track is not muted, solo the requested track
        for requested_solo_track in requested_solo_tracks:
            if not requested_solo_track.track.mute:
                self.should_solo(self.looper_tracks.values(), exclusive_solo)
                requested_solo_track.track.solo = True

        #check for mute all, solo all non-datalooper tracks and non-group tracks
        if not mute_all:
            for track in self.tracks:
                if track not in self.group_tracks and not ("DL#" in track.name or "CL#" in track.name):
                    track.solo = True

    def should_solo(self, looper_track_lists, exclusive_solo):
        for looper_track_list in looper_track_lists:
            for looper_track in looper_track_list:
                # if there's a CL Track that is not playing, let it be solo'd so you can still pass audio through -- i. e. a live instrument
                if (exclusive_solo or not looper_track.track.solo):
                    looper_track.track.solo = False
                else:
                    looper_track.track.solo = True

    def unsolo_track(self, requested_solo_tracks):
        self.__parent.send_message("unsoloing tracks")
        for requested_solo_track in requested_solo_tracks:
            requested_solo_track.track.solo = False

        unsolo_all = True

        for looper_track_list in self.looper_tracks.values():
            for looper_track in looper_track_list:
                if looper_track.track.solo:
                    unsolo_all = False

        if unsolo_all:
            for track in self.tracks:
                track.solo = False

    def link_button(self, button_number, track_number):
        if track_number >= 0:
            self.linked_buttons[track_number] = Solo.SoloTrack(self.__parent, button_number, self.looper_tracks.get(str(track_number)))
            self.__parent.send_message("linking button number: " + str(button_number) + " to track number: " + str(track_number) + " solo control")

    class SoloTrack:
        def __init__(self, __parent, button_number, looper_tracks):
            #tracks corresponding to a single #, CL#1 1, CL#1 2
            self.looper_tracks = looper_tracks
            self.__parent = __parent
            self.button_number = button_number
            if looper_tracks:
                for looper_track in looper_tracks:
                        looper_track.track.add_solo_listener(self.on_solo_change)
                        self.on_solo_change()

        def on_solo_change(self):
            light_on = False
            for looper_track in self.looper_tracks:
                if looper_track.track.solo:
                    light_on = True
            if light_on:
                carray = self.__parent.get_color_array(1262335)
                self.__parent.send_message("Requesting color change on button #" + str(self.button_number))
                self.__parent.send_sysex(DLCommands.CHANGE_COLOR, self.button_number, carray[0], carray[1], carray[2], carray[3], carray[4], carray[5], False)
            else:
                self.__parent.send_sysex(DLCommands.RESTORE_COLOR, self.button_number)

