from .consts import *
import math
import Live
from .state import *

class Track(object):
    """ Class handling looper track from DataLooper """

    def __init__(self, parent, track, trackNum, song, state, action_handler):
        self.__parent = parent
        self.track = track
        self.trackNum = trackNum
        self.song = song
        self.led_disabled = False
        self.global_state = state
        self.action_handler = action_handler
        self.clearTimerCounter = 0
        self.gain_parameter = -1
        self.stopGainParam = -1
        self.volume_initial = 0
        self.preStopGain = 0
        self.fade_timer = -1
        self.undo_button = -1
        self.clear_ignoring_arm = False
        self.group_number = -1
        self.group_master = -1
        if self.track.can_be_armed:
            self.orig_arm = self.track.arm
        else:
            self.orig_arm = 1
        self.lastState = CLEAR_STATE
        self.button_num = -1
        self.mute_button = -1
        self.should_toggle = False
        self.should_clear_all = False
        self.track.add_mute_listener(self.on_mute)

        # Beat Delay Variables
        self.rec_delay_timer = Live.Base.Timer(self.rec_delay, interval=1, repeat=True)
        self.fixed_rec_timer = Live.Base.Timer(self.fixed_length_rec, interval=1, repeat=True)
        self.group_toggle_timer = Live.Base.Timer(self.toggle_groups, interval=1, repeat=False)



    def clear_listener(self):
        pass

    def toggle_playback(self):
        pass

    def send_message(self, message):
        self.__parent.send_message(message)

    def overdub(self):
        if self.lastState == PLAYING_STATE:
            self.trigger_record()

    def trigger_record(self):
        pass

    def undo(self):
        pass

    def fadeToMS(self, fadeTime):
        return (fadeTime / 127.) * 31500
    def clear(self, fadeTime = 0):
        self.send_message("fade time: " + str(fadeTime) + " clear timer is false? : " + str(self.fade_timer))
        self.start_fade(fadeTime, self.clear_fade_complete, self.clear_immediately)

    def clear_ignoring_rec_arm(self, fadeTime = 0):
        self.clear_ignoring_arm = True
        self.clear(fadeTime)
        self.clear_ignoring_arm = False
        
    def start_fade(self, fadeTime, on_complete_method, no_fade_method, arg = -1):
        self.send_message("in start fade")
        if fadeTime > 0 and self.fade_timer == -1:
            fadeTime = self.fadeToMS(fadeTime)
            self.send_message("starting fade")
            self.send_message("volume: " + str(self.track.mixer_device.volume.value))
            self.volume_initial = self.track.mixer_device.volume.value
            self.fade_timer = Live.Base.Timer(callback=lambda: self.fade(self.volume_initial * (10 / fadeTime), self.fade_timer, self.volume_initial, on_complete_method), interval=10, repeat=True)
            self.fade_timer.start()
        else:
            self.send_message("no fade, calling method: " + str(no_fade_method))
            if arg != -1:
                no_fade_method(arg)
            else:
                no_fade_method()

    def fade(self, increment, timer, firstValue, methodOnComplete):
        # self.send_message("gain min:" + str(self.parameter.min))
        volume = self.track.mixer_device.volume.value
        if volume > increment:
            self.track.mixer_device.volume.value = self.track.mixer_device.volume.value - increment
        else:
            self.send_message("clearing timer in first else")
            timer.stop()
            self.reset_gain(firstValue, methodOnComplete)


    def clear_immediately(self):
        if self.group_master != -1:
            self.action_handler.toggle_groups(self.group_master, False)
        self.send_message("calling clear immediately")
        if len(self.track.devices) and self.track.devices[0].can_have_chains and not self.track.devices[0].can_have_drum_pads and hasattr(self.track.devices[0],'variation_count') and self.track.devices[0].variation_count > 0:
            self.send_message("variation count" + str(self.track.devices[0].variation_count) )
            self.track.devices[0].selected_variation_index = 0
            self.track.devices[0].recall_selected_variation()
            self.track.mute = 0
            if "VOL" in self.track.name:
                self.send_message("Resetting volume on track: " + str(self.track.name))
                self.track.mixer_device.volume.value = .85 # 0db
        if self.fade_timer != -1:
            self.fade_timer.stop()
            self.track.mixer_device.volume.value = self.volume_initial
            self.fade_timer = -1

    def reset_gain(self, firstValue, methodOnComplete):
        self.track.mixer_device.volume.value = firstValue
        methodOnComplete()

    def start(self, start_type):
        if self.lastState != CLEAR_STATE:
            self.play(start_type)

    def record(self, quantized):
        pass
    def play(self,quantized):
        pass
    def on_quantize_disabled(self):
        pass

    def set_group(self, group_num = [-1]):
        self.group_number = group_num[0]

    def set_group_master(self, group_num = -1):
        if group_num != -1:
            self.send_message("setting group master to num: ")
            for num in group_num:
                self.send_message(str(num) + " ")
        self.group_master = group_num

    def set_overdub(self):
        pass
    def set_led(self):
        pass

    def update_state(self, state):

        if self.track and state != -1 :
            # self.send_message("current monitoring state: " + str(self.track.current_monitoring_state))

            # if state has actually changed, and the button is linked to the current preset, change to slow blink
            if self.lastState != state and self.global_state.mode != NEW_SESSION_MODE and self.button_num is not -1:
                self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.SLOW_BLINK)
            trackInReturn = self.track in self.song.return_tracks
            trackInArray = len(self.action_handler.tracks.get(str(self.trackNum))) >= 1
            if not self.led_disabled and (trackInReturn or trackInArray):
                button_val = self.button_num if self.button_num is not -1 else 13
                self.__parent.send_sysex(DLCommands.CHANGE_STATE, button_val, state,self.trackNum)
                #this prevents DL script from sending state to other controllers if a DL pedal is not connected. It was added because stop state was being sent on clear from non-connected pedals
                if button_val is not 13 or (button_val is 13 and state != STOP_STATE):
                    self.__parent.send_looper_state(self.trackNum, state)
                self.send_message("updating state on track num: " + str(self.trackNum) + " from state: " + str(self.lastState) + " to: " + str(state) + " track name: " + self.track.name + " button value:"  + str(button_val))
            else:
                self.send_message("no update sent, state updated internally to " + str(state) + " on track " + self.track.name + " LED Disabled? " + str(self.led_disabled) + " button num:" + str(self.button_num))

            if self.lastState == PLAYING_STATE or state == OVERDUB_STATE:
                self.global_state.mode = LOOPER_MODE
            elif self.lastState == RECORDING_STATE and state == PLAYING_STATE and self.should_toggle:
                pass
                ## self.should_toggle = False
                ## self.group_toggle_timer.start()
            self.lastState = state
        # if self.trackNum not in self.__parent.duplicates or (
        #         self.trackNum in self.__parent.duplicates and "LED" in self.track.name):
        #     self.send_sysex(self.trackNum, DLCommands.CHANGE_STATE, self.lastState)
    
    def toggle_groups(self):
        self.action_handler.toggle_groups(self.group_master, True)

    def request_state(self):
        self.send_message("sending requested state on track num: " + str(self.trackNum) + " to state: " + str(self.lastState) + " track name: " + self.track.name + " button number:" + str(self.button_num))
        if self.button_num != -1 and not self.led_disabled and (not self.track.can_be_armed or self.track.arm or self.track.current_monitoring_state is 0):
            self.__parent.send_sysex(DLCommands.CHANGE_STATE, self.button_num, self.lastState)
        if self.mute_button != -1:
            if self.track.mute:
                mute_state = 5
            else:
                mute_state = 4
            self.__parent.send_sysex(DLCommands.CHANGE_STATE, self.mute_button, mute_state)

    def disable_led(self):
        self.send_message("disabling led on track name:" + str(self.track.name))
        self.led_disabled = True

    def remove_track(self):
        pass

    def change_mode(self):
        pass

    def record_quantized(self, action_during_play = 0):
        if self.lastState == PLAYING_STATE  and action_during_play > 0:
            getattr(self,LOOPER_ACTIONS.get(action_during_play))()
        elif self.song.is_playing and self.lastState == CLEAR_STATE and RecBeatDelay.enabled:
            RecBeatDelay.num_beats_when_triggered = round(self.song.current_song_time)
            self.rec_delay_timer.start()
            self.__parent.send_sysex(DLCommands.BLINK, self.button_num, BlinkTypes.FAST_BLINK)
            if not RecBeatDelay.should_persist:
                self.send_message("Disabling record delay. Sending SYSEX status update")
                RecBeatDelay.enabled = 0
                self.__parent.send_sysex(REC_DELAY_TOGGLED, 0)
        else:
            if self.rec_delay_timer.running:
                self.rec_delay_timer.stop()
            self.check_for_fixed_length_rec()
            self.record(True)
        if self.lastState == RECORDING_STATE and self.group_master != -1:
            self.send_message("attempting group mute, adding time change listener")
            self.should_toggle = True
            self.song.add_current_song_time_listener(self.on_time_change)

    def on_time_change(self):
        time = self.song.get_current_beats_song_time()

        if self.should_toggle and time.sub_division == 4:
            self.should_toggle == False
            self.song.remove_current_song_time_listener(self.on_time_change)
            self.group_toggle_timer.start()

            
    def check_for_fixed_length_rec(self):
        if FixedLengthLoop.enabled and self.lastState == CLEAR_STATE and self.song.is_playing:
            FixedLengthLoop.num_beats_when_triggered = math.ceil(self.song.current_song_time)
            self.fixed_rec_timer.start()
            if not FixedLengthLoop.should_persist:
                self.send_message("Disabling fixed length loop. Sending SYSEX status update")
                FixedLengthLoop.enabled = 0
                self.__parent.send_sysex(FIXED_LENGTH_REC_TOGGLED, 0)


    def rec_delay(self):
        if RecBeatDelay.num_beats_when_triggered + RecBeatDelay.num_beats_delayed - .5  <= self.song.current_song_time:
            self.rec_delay_timer.stop()
            self.check_for_fixed_length_rec()
            self.record(True)


    def fixed_length_rec(self):
        if FixedLengthLoop.num_beats_when_triggered + FixedLengthLoop.num_beats_to_loop - .5 <= self.song.current_song_time:
            self.fixed_rec_timer.stop()
            self.record(True)

    def record_immediately(self, action_during_play = 0):
        if self.lastState == PLAYING_STATE  and action_during_play > 0:
            getattr(self,LOOPER_ACTIONS.get(action_during_play))()
        else:
            self.record(False)

    def record_ignoring_state(self):
        pass

    def stop(self, quantized):
        pass

    def exclusive_record_quantized(self, looperNum = -1):
        self.exclusive_record(True, looperNum)

    def exclusive_record_immediately(self, looperNum = -1):
        self.exclusive_record(False, looperNum)

    def exclusive_record(self, quantized = True, looperNum = -1):
        pass

    def stop_quantized(self, fadeTime = 0):
        self.send_message("stop quantized called, fade time: "+ str(fadeTime))
        self.start_fade(fadeTime, self.stop_fade_complete, self.stop, True)

    def stop_immediately(self):
        pass

    def stop_fade_complete(self):
        self.fade_timer = -1
        self.stop(False)

    def clear_fade_complete(self):
        self.fade_timer = -1
        self.clear_immediately()

    def link_button(self, button_num, action):
        if action == 'record_immediately' or action == 'record_quantized' or action == 'record_unquantized' or action == 'exclusive_record_quantized' or action == 'exclusive_record_unquantized':
            self.send_message("linking button: " + str(button_num))
            self.button_num = button_num
            self.request_state()
        elif action == 'toggle_mute':
            self.mute_button = button_num
            self.request_state()
        elif action == 'undo':
            self.undo_button = button_num

    def toggle_mute(self):
        self.track.mute = not self.track.mute

    def on_mute(self):
        self.send_message("track mute changed")
        self.request_state()

    def remove_clip_slot(self):
        pass

    def none(self, action_during_play = 0):
        self.send_message("none called, attempting action during play:" + str(action_during_play))
        if self.lastState == PLAYING_STATE and action_during_play > 0:
            self.send_message("none called, calling action during play:" + str(LOOPER_ACTIONS.get(action_during_play)))
            getattr(self, LOOPER_ACTIONS.get(action_during_play))()

